import datetime
import logging
import os
import time
import traceback
from typing import Any, Dict, List, Optional, Union

import torch
import torch.distributed as dist
import torch.optim as optim
import torch.optim.lr_scheduler
import torch.utils.data as data
from allennlp import training, common
from allennlp.common import checks
from allennlp.common import util as common_util
from allennlp.models import model
from allennlp.training import checkpointer, optimizers
from allennlp.training import learning_rate_schedulers
from allennlp.training import momentum_schedulers
from allennlp.training import moving_average
from allennlp.training import tensorboard_writer as allen_tensorboard_writer
from allennlp.training import util as training_util
from overrides import overrides

from unidic_combo.training import tensorboard_writer as combo_tensorboard_writer

logger = logging.getLogger(__name__)


@training.EpochCallback.register("transfer_patience")
class TransferPatienceEpochCallback(training.EpochCallback):

    def __call__(self, trainer: "training.GradientDescentTrainer", metrics: Dict[str, Any], epoch: int,
                 is_master: bool) -> None:
        if trainer._learning_rate_scheduler and trainer._learning_rate_scheduler.patience is not None:
            trainer._metric_tracker._patience = trainer._learning_rate_scheduler.patience
            trainer._metric_tracker._epochs_with_no_improvement = 0
        else:
            raise checks.ConfigurationError("Learning rate scheduler isn't properly setup!")


@training.Trainer.register("gradient_descent_validate_n", constructor="from_partial_objects")
class GradientDescentTrainer(training.GradientDescentTrainer):

    def __init__(self, model: model.Model, optimizer: optim.Optimizer, data_loader: data.DataLoader,
                 patience: Optional[int] = None, validation_metric: str = "-loss",
                 validation_data_loader: data.DataLoader = None, num_epochs: int = 20,
                 serialization_dir: Optional[str] = None, checkpointer: checkpointer.Checkpointer = None,
                 cuda_device: Optional[Union[int, torch.device]] = -1,
                 grad_norm: Optional[float] = None, grad_clipping: Optional[float] = None,
                 learning_rate_scheduler: Optional[learning_rate_schedulers.LearningRateScheduler] = None,
                 momentum_scheduler: Optional[momentum_schedulers.MomentumScheduler] = None,
                 tensorboard_writer: allen_tensorboard_writer.TensorboardWriter = None,
                 moving_average: Optional[moving_average.MovingAverage] = None,
                 batch_callbacks: List[training.BatchCallback] = None,
                 epoch_callbacks: List[training.EpochCallback] = None,
                 end_callbacks: List[training.EpochCallback] = None,
                 trainer_callbacks: List[training.TrainerCallback] = None,
                 distributed: bool = False, local_rank: int = 0,
                 world_size: int = 1, num_gradient_accumulation_steps: int = 1,
                 use_amp: bool = False) -> None:
        super().__init__(model, optimizer, data_loader, patience, validation_metric, validation_data_loader, num_epochs,
                         serialization_dir, checkpointer, cuda_device, grad_norm, grad_clipping,
                         learning_rate_scheduler, momentum_scheduler, tensorboard_writer, moving_average,
                         batch_callbacks, epoch_callbacks, end_callbacks, trainer_callbacks, distributed, local_rank, world_size,
                         num_gradient_accumulation_steps, use_amp)
        # TODO extract param to constructor (+ constructor method?)
        self.validate_every_n = 5

    @overrides
    def train(self) -> Dict[str, Any]:
        """
        Trains the supplied model with the supplied parameters.
        """
        try:
            epoch_counter = self._restore_checkpoint()
        except RuntimeError:
            traceback.print_exc()
            raise checks.ConfigurationError(
                "Could not recover training from the checkpoint.  Did you mean to output to "
                "a different serialization directory or delete the existing serialization "
                "directory?"
            )

        training_util.enable_gradient_clipping(self.model, self._grad_clipping)

        logger.info("Beginning training.")

        val_metrics: Dict[str, float] = {}
        this_epoch_val_metric: float = None
        metrics: Dict[str, Any] = {}
        epochs_trained = 0
        training_start_time = time.time()

        metrics["best_epoch"] = self._metric_tracker.best_epoch
        for key, value in self._metric_tracker.best_epoch_metrics.items():
            metrics["best_validation_" + key] = value

        for callback in self._epoch_callbacks:
            callback(self, metrics={}, epoch=-1, is_master=True)

        for epoch in range(epoch_counter, self._num_epochs):
            epoch_start_time = time.time()
            train_metrics = self._train_epoch(epoch)

            # get peak of memory usage
            if "cpu_memory_MB" in train_metrics:
                metrics["peak_cpu_memory_MB"] = max(
                    metrics.get("peak_cpu_memory_MB", 0), train_metrics["cpu_memory_MB"]
                )
            for key, value in train_metrics.items():
                if key.startswith("gpu_"):
                    metrics["peak_" + key] = max(metrics.get("peak_" + key, 0), value)

            if self._validation_data_loader is not None:
                val_metrics = {}
                this_epoch_val_metric = None
                if epoch % self.validate_every_n == 0:
                    with torch.no_grad():
                        # We have a validation set, so compute all the metrics on it.
                        val_loss, val_reg_loss, num_batches = self._validation_loss(epoch)

                        # It is safe again to wait till the validation is done. This is
                        # important to get the metrics right.
                        if self._distributed:
                            dist.barrier()

                        val_metrics = training_util.get_metrics(
                            self.model,
                            val_loss,
                            val_reg_loss,
                            num_batches=num_batches,
                            batch_loss=None,
                            batch_reg_loss=None,
                            reset=True,
                            world_size=self._world_size,
                            cuda_device=self.cuda_device,
                        )

                        # Check validation metric for early stopping
                        this_epoch_val_metric = val_metrics[self._validation_metric]
                        # self._metric_tracker.add_metric(this_epoch_val_metric)

                train_metrics["patience"] = self._metric_tracker._patience
                if self._metric_tracker.should_stop_early():
                    logger.info("Ran out of patience.  Stopping training.")
                    break

            if self._master:
                self._tensorboard.log_metrics(
                    train_metrics, val_metrics=val_metrics, log_to_console=True, epoch=epoch + 1
                )  # +1 because tensorboard doesn't like 0

            # Create overall metrics dict
            training_elapsed_time = time.time() - training_start_time
            metrics["training_duration"] = str(datetime.timedelta(seconds=training_elapsed_time))
            metrics["training_start_epoch"] = epoch_counter
            metrics["training_epochs"] = epochs_trained
            metrics["epoch"] = epoch

            for key, value in train_metrics.items():
                metrics["training_" + key] = value
            for key, value in val_metrics.items():
                metrics["validation_" + key] = value

            if self._metric_tracker.is_best_so_far():
                # Update all the best_ metrics.
                # (Otherwise they just stay the same as they were.)
                metrics["best_epoch"] = epoch
                for key, value in val_metrics.items():
                    metrics["best_validation_" + key] = value

                self._metric_tracker.best_epoch_metrics = val_metrics

            if self._serialization_dir and self._master:
                common_util.dump_metrics(
                    os.path.join(self._serialization_dir, f"metrics_epoch_{epoch}.json"), metrics
                )

            # The Scheduler API is agnostic to whether your schedule requires a validation metric -
            # if it doesn't, the validation metric passed here is ignored.
            if self._learning_rate_scheduler:
                self._learning_rate_scheduler.step(this_epoch_val_metric)
            if self._momentum_scheduler:
                self._momentum_scheduler.step(this_epoch_val_metric)

            if self._master:
                self._checkpointer.save_checkpoint(
                    epoch, self, is_best_so_far=self._metric_tracker.is_best_so_far()
                )

            # Wait for the master to finish saving the checkpoint
            if self._distributed:
                dist.barrier()

            for callback in self._epoch_callbacks:
                callback(self, metrics=metrics, epoch=epoch, is_master=self._master)

            epoch_elapsed_time = time.time() - epoch_start_time
            logger.info("Epoch duration: %s", datetime.timedelta(seconds=epoch_elapsed_time))

            if epoch < self._num_epochs - 1:
                training_elapsed_time = time.time() - training_start_time
                estimated_time_remaining = training_elapsed_time * (
                        (self._num_epochs - epoch_counter) / float(epoch - epoch_counter + 1) - 1
                )
                formatted_time = str(datetime.timedelta(seconds=int(estimated_time_remaining)))
                logger.info("Estimated training time remaining: %s", formatted_time)

            epochs_trained += 1

        # make sure pending events are flushed to disk and files are closed properly
        self._tensorboard.close()

        # Load the best model state before returning
        best_model_state = self._checkpointer.best_model_state()
        if best_model_state:
            self.model.load_state_dict(best_model_state)

        return metrics

    @classmethod
    def from_partial_objects(
            cls,
            model: model.Model,
            serialization_dir: str,
            data_loader: data.DataLoader,
            validation_data_loader: data.DataLoader = None,
            local_rank: int = 0,
            patience: int = None,
            validation_metric: str = "-loss",
            num_epochs: int = 20,
            cuda_device: int = -1,
            grad_norm: float = None,
            grad_clipping: float = None,
            distributed: bool = None,
            world_size: int = 1,
            num_gradient_accumulation_steps: int = 1,
            opt_level: Optional[str] = None,
            use_amp: bool = False,
            optimizer: common.Lazy[optimizers.Optimizer] = None,
            learning_rate_scheduler: common.Lazy[learning_rate_schedulers.LearningRateScheduler] = None,
            momentum_scheduler: common.Lazy[momentum_schedulers.MomentumScheduler] = None,
            tensorboard_writer: common.Lazy[allen_tensorboard_writer.TensorboardWriter] = None,
            moving_average: common.Lazy[moving_average.MovingAverage] = None,
            checkpointer: common.Lazy[training.Checkpointer] = None,
            batch_callbacks: List[training.BatchCallback] = None,
            epoch_callbacks: List[training.EpochCallback] = None,
    ) -> "training.Trainer":
        if tensorboard_writer is None:
            tensorboard_writer = common.Lazy(combo_tensorboard_writer.NullTensorboardWriter)
        return super().from_partial_objects(
            model=model,
            serialization_dir=serialization_dir,
            data_loader=data_loader,
            validation_data_loader=validation_data_loader,
            local_rank=local_rank,
            patience=patience,
            validation_metric=validation_metric,
            num_epochs=num_epochs,
            cuda_device=cuda_device,
            grad_norm=grad_norm,
            grad_clipping=grad_clipping,
            distributed=distributed,
            world_size=world_size,
            num_gradient_accumulation_steps=num_gradient_accumulation_steps,
            use_amp=use_amp,
            optimizer=optimizer,
            learning_rate_scheduler=learning_rate_scheduler,
            momentum_scheduler=momentum_scheduler,
            tensorboard_writer=tensorboard_writer,
            moving_average=moving_average,
            checkpointer=checkpointer,
            batch_callbacks=batch_callbacks,
            epoch_callbacks=epoch_callbacks,
        )



"""Models module."""
from .base import FeedForwardPredictor
from .parser import DependencyRelationModel
from .embeddings import CharacterBasedWordEmbeddings
from .encoder import ComboEncoder
from .lemma import LemmatizerModel
from .model import SemanticMultitaskModel
from .morpho import MorphologicalFeatures



from typing import Dict, Optional, List, Union

import torch
import torch.nn as nn
from allennlp import common, data
from allennlp import nn as allen_nn
from allennlp.common import checks
from allennlp.modules import feedforward
from allennlp.nn import Activation

from unidic_combo.models import utils


class Predictor(nn.Module, common.Registrable):
    default_implementation = "feedforward_predictor_from_vocab"

    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()


class Linear(nn.Linear, common.FromParams):

    def __init__(self,
                 in_features: int,
                 out_features: int,
                 activation: Optional[allen_nn.Activation] = lambda x: x,
                 dropout_rate: Optional[float] = 0.0):
        super().__init__(in_features, out_features)
        self.activation = activation
        self.dropout = nn.Dropout(p=dropout_rate) if dropout_rate else lambda x: x

    def forward(self, x: torch.FloatTensor) -> torch.FloatTensor:
        x = super().forward(x)
        x = self.activation(x)
        return self.dropout(x)

    def get_output_dim(self) -> int:
        return self.out_features


@Predictor.register("feedforward_predictor")
@Predictor.register("feedforward_predictor_from_vocab", constructor="from_vocab")
class FeedForwardPredictor(Predictor):
    """Feedforward predictor. Should be used on top of Seq2Seq encoder."""

    def __init__(self, feedforward_network: feedforward.FeedForward):
        super().__init__()
        self.feedforward_network = feedforward_network

    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        if mask is None:
            mask = x.new_ones(x.size()[:-1])

        x = self.feedforward_network(x)
        output = {
            "prediction": x.argmax(-1),
            "probability": x
        }

        if labels is not None:
            if sample_weights is None:
                sample_weights = labels.new_ones([mask.size(0)])
            output["loss"] = self._loss(x, labels, mask, sample_weights)

        return output

    def _loss(self,
              pred: torch.Tensor,
              true: torch.Tensor,
              mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> torch.Tensor:
        BATCH_SIZE, _, CLASSES = pred.size()
        valid_positions = mask.sum()
        pred = pred.reshape(-1, CLASSES)
        true = true.reshape(-1)
        mask = mask.reshape(-1)
        loss = utils.masked_cross_entropy(pred, true, mask)
        loss = loss.reshape(BATCH_SIZE, -1) * sample_weights.unsqueeze(-1)
        return loss.sum() / valid_positions

    @classmethod
    def from_vocab(cls,
                   vocab: data.Vocabulary,
                   vocab_namespace: str,
                   input_dim: int,
                   num_layers: int,
                   hidden_dims: List[int],
                   activations: Union[Activation, List[Activation]],
                   dropout: Union[float, List[float]] = 0.0,
                   ):
        if len(hidden_dims) + 1 != num_layers:
            raise checks.ConfigurationError(
                f"len(hidden_dims) ({len(hidden_dims):d}) + 1 != num_layers ({num_layers:d})"
            )

        assert vocab_namespace in vocab.get_namespaces(), \
            f"There is not {vocab_namespace} in created vocabs, check if this field has any values to predict!"
        hidden_dims = hidden_dims + [vocab.get_vocab_size(vocab_namespace)]

        return cls(feedforward.FeedForward(
            input_dim=input_dim,
            num_layers=num_layers,
            hidden_dims=hidden_dims,
            activations=activations,
            dropout=dropout))



from typing import List

import torch
import torch.nn as nn

from allennlp import common
from allennlp import nn as allen_nn


class DilatedCnnEncoder(nn.Module, common.FromParams):

    def __init__(self,
                 input_dim: int,
                 filters: List[int],
                 kernel_size: List[int],
                 stride: List[int],
                 padding: List[int],
                 dilation: List[int],
                 activations: List[allen_nn.Activation]):
        super().__init__()
        conv1d_layers = []
        input_dims = [input_dim] + filters[:-1]
        output_dims = filters
        for idx in range(len(activations)):
            conv1d_layers.append(nn.Conv1d(
                in_channels=input_dims[idx],
                out_channels=output_dims[idx],
                kernel_size=kernel_size[idx],
                stride=stride[idx],
                padding=padding[idx],
                dilation=dilation[idx]))
        self.conv1d_layers = nn.ModuleList(conv1d_layers)
        self.activations = activations
        assert len(self.activations) == len(self.conv1d_layers)

    def forward(self, x: torch.FloatTensor) -> torch.FloatTensor:
        for layer, activation in zip(self.conv1d_layers, self.activations):
            x = activation(layer(x))
        return x



#! /usr/bin/python3 -i
# coding=utf-8

import os
PACKAGE_DIR=os.path.abspath(os.path.dirname(__file__))
combo_parser=None

def ComboAPI(conllu):
  from unidic_combo.data import Token,Sentence,sentence2conllu
  d=[]
  e=[]
  for s in conllu.split("\n"):
    if s=="" or s.startswith("#"):
      if e!=[]:
        d.append(Sentence(tokens=e))
        e=[]
    else:
      t=s.split("\t")
      e.append(Token(id=int(t[0]),token=t[1],lemma=t[2],upostag=t[3],xpostag=t[4],misc=t[9]))
  u=combo_parser(d)
  for s in u:
    for t in s.tokens:
      d=t.deprel
      if d=="root":
        if t.head!=0:
          t.deprel="advcl" if t.head>t.id else "parataxis"
      elif d=="advmod":
        t.upostag="ADV"
      elif d=="amod":
        t.upostag="ADJ"
      elif d=="aux" or d=="cop":
        t.upostag="AUX"
      elif d=="det":
        t.upostag="DET"
      elif d=="nummod":
        t.upostag="NUM"
      if t.head==0 or t.head==t.id:
        t.head=0
        t.deprel="root"
  return "".join([sentence2conllu(s,False).serialize() for s in u])

def ComboRevAPI(conllu):
  from unidic_combo.data import Token,Sentence,sentence2conllu
  d=[]
  e=[]
  for s in conllu.split("\n"):
    if s=="" or s.startswith("#"):
      if e!=[]:
        d.append(Sentence(tokens=e))
        e=[]
    else:
      t=s.split("\t")
      e.append(Token(id=int(t[0]),token=t[2],lemma=t[1],upostag=t[3],xpostag=t[4],misc=t[9]))
  u=combo_parser(d)
  for s in u:
    for t in s.tokens:
      t.token,t.lemma=t.lemma,t.token
      d=t.deprel
      if d=="root":
        if t.head!=0:
          t.deprel="advcl" if t.head>t.id else "parataxis"
      elif d=="advmod":
        t.upostag="ADV"
      elif d=="amod":
        t.upostag="ADJ"
      elif d=="aux" or d=="cop":
        t.upostag="AUX"
      elif d=="det":
        t.upostag="DET"
      elif d=="nummod":
        t.upostag="NUM"
      if t.head==0 or t.head==t.id:
        t.head=0
        t.deprel="root"
  return "".join([sentence2conllu(s,False).serialize() for s in u])

def load(UniDic=None,BERT=True,LemmaAsForm=None):
  global combo_parser
  import unidic2ud.spacy
  if UniDic==None:
    UniDic="ipadic"
  nlp=unidic2ud.spacy.load(UniDic,None)
  m="combo-japanese.tar.gz" if BERT else "combo-japanese-small.tar.gz"
  if LemmaAsForm==None:
    LemmaAsForm=UniDic not in ["gendai","spoken","ipadic"]
  if LemmaAsForm:
    m=m.replace(".tar.gz","-rev.tar.gz")
  if nlp.tokenizer.model.model!=m:
    combo_parser=None
  if combo_parser==None:
    import unidic_combo.predict
    combo_parser=unidic_combo.predict.SemanticMultitaskPredictor.from_pretrained(os.path.join(PACKAGE_DIR,m))
    nlp.tokenizer.model.udpipe=ComboRevAPI if LemmaAsForm else ComboAPI
  nlp.tokenizer.model.model=m
  return nlp




import logging
from typing import Union, List, Dict, Iterable, Optional, Any

import conllu
from allennlp import data as allen_data
from allennlp.common import checks
from allennlp.data import fields as allen_fields, vocabulary
from conllu import parser
from dataclasses import dataclass
from overrides import overrides

from unidic_combo.data import fields

logger = logging.getLogger(__name__)


@allen_data.DatasetReader.register("conllu")
class UniversalDependenciesDatasetReader(allen_data.DatasetReader):

    def __init__(
            self,
            token_indexers: Dict[str, allen_data.TokenIndexer] = None,
            lemma_indexers: Dict[str, allen_data.TokenIndexer] = None,
            features: List[str] = None,
            targets: List[str] = None,
            use_sem: bool = False,
            **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if features is None:
            features = ["token", "char"]
        if targets is None:
            targets = ["head", "deprel", "upostag", "xpostag", "lemma", "feats"]

        if "token" not in features and "char" not in features:
            raise checks.ConfigurationError("There must be at least one ('char' or 'token') text-based feature!")

        intersection = set(features).intersection(set(targets))
        if len(intersection) != 0:
            raise checks.ConfigurationError(
                "Features and targets cannot share elements! "
                "Remove {} from either features or targets.".format(intersection)
            )
        self.use_sem = use_sem

        # *.conllu readers configuration
        fields = list(parser.DEFAULT_FIELDS)
        fields[1] = "token"  # use 'token' instead of 'form'
        field_parsers = parser.DEFAULT_FIELD_PARSERS
        # Do not make it nullable
        field_parsers.pop("xpostag", None)
        # Ignore parsing misc
        field_parsers.pop("misc", None)
        if self.use_sem:
            fields = list(fields)
            fields.append("semrel")
            field_parsers["semrel"] = lambda line, i: line[i]
        self.field_parsers = field_parsers
        self.fields = tuple(fields)

        self._token_indexers = token_indexers
        self._lemma_indexers = lemma_indexers
        self._targets = targets
        self._features = features
        self.generate_labels = True
        # Filter out not required token indexers to avoid
        # Mismatched token keys ConfigurationError
        for indexer_name in list(self._token_indexers.keys()):
            if indexer_name not in self._features:
                del self._token_indexers[indexer_name]

    @overrides
    def _read(self, file_path: str) -> Iterable[allen_data.Instance]:
        file_path = [file_path] if len(file_path.split(",")) == 0 else file_path.split(",")

        for conllu_file in file_path:
            with open(conllu_file, "r") as file:
                for annotation in conllu.parse_incr(file, fields=self.fields, field_parsers=self.field_parsers):
                    yield self.text_to_instance(annotation)

    @overrides
    def text_to_instance(self, tree: conllu.TokenList) -> allen_data.Instance:
        fields_: Dict[str, allen_data.Field] = {}
        tree_tokens = [t for t in tree if isinstance(t["id"], int)]
        tokens = [_Token(t["token"],
                         pos_=t.get("upostag"),
                         tag_=t.get("xpostag"),
                         lemma_=t.get("lemma"),
                         feats_=t.get("feats"))
                  for t in tree_tokens]

        # features
        text_field = allen_fields.TextField(tokens, self._token_indexers)
        fields_["sentence"] = text_field

        # targets
        if self.generate_labels:
            for target_name in self._targets:
                if target_name != "sent":
                    target_values = [t[target_name] for t in tree_tokens]
                    if target_name == "lemma":
                        target_values = [allen_data.Token(v) for v in target_values]
                        fields_[target_name] = allen_fields.TextField(target_values, self._lemma_indexers)
                    elif target_name == "feats":
                        target_values = self._feat_values(tree_tokens)
                        fields_[target_name] = fields.SequenceMultiLabelField(target_values,
                                                                              self._feats_to_index_multi_label,
                                                                              text_field,
                                                                              label_namespace="feats_labels")
                    elif target_name == "head":
                        target_values = [0 if v == "_" else int(v) for v in target_values]
                        fields_[target_name] = allen_fields.SequenceLabelField(target_values, text_field,
                                                                               label_namespace=target_name + "_labels")
                    else:
                        fields_[target_name] = allen_fields.SequenceLabelField(target_values, text_field,
                                                                               label_namespace=target_name + "_labels")

        # Restore feats fields to string representation
        # parser.serialize_field doesn't handle key without value
        for token in tree.tokens:
            if "feats" in token:
                feats = token["feats"]
                if feats:
                    feats_values = []
                    for k, v in feats.items():
                        feats_values.append('='.join((k, v)) if v else k)
                    field = "|".join(feats_values)
                else:
                    field = "_"
                token["feats"] = field

        # metadata
        fields_["metadata"] = allen_fields.MetadataField({"input": tree, "field_names": self.fields})

        return allen_data.Instance(fields_)

    @staticmethod
    def _feat_values(tree: List[Dict[str, Any]]):
        features = []
        for token in tree:
            token_features = []
            if token["feats"] is not None:
                for feat, value in token["feats"].items():
                    if feat in ["_", "__ROOT__"]:
                        pass
                    else:
                        # Handle case where feature is binary (doesn't have associated value)
                        if value:
                            token_features.append(feat + "=" + value)
                        else:
                            token_features.append(feat)
            features.append(token_features)
        return features

    @staticmethod
    def _feats_to_index_multi_label(vocab: allen_data.Vocabulary):
        label_namespace = "feats_labels"
        vocab_size = vocab.get_vocab_size(label_namespace)
        slices = get_slices_if_not_provided(vocab)

        def _m_from_n_ones_encoding(multi_label: List[str]) -> List[int]:
            one_hot_encoding = [0] * vocab_size
            for cat, cat_indices in slices.items():
                if cat not in ["__PAD__", "_"]:
                    label_from_cat = [label for label in multi_label if cat == label.split("=")[0]]
                    if label_from_cat:
                        label_from_cat = label_from_cat[0]
                        index = vocab.get_token_index(label_from_cat, label_namespace)
                    else:
                        # Get Cat=None index
                        index = vocab.get_token_index(cat + "=None", label_namespace)
                    one_hot_encoding[index] = 1
            return one_hot_encoding

        return _m_from_n_ones_encoding


@allen_data.Vocabulary.register("from_instances_extended", constructor="from_instances_extended")
class Vocabulary(allen_data.Vocabulary):

    @classmethod
    def from_instances_extended(
            cls,
            instances: Iterable[allen_data.Instance],
            min_count: Dict[str, int] = None,
            max_vocab_size: Union[int, Dict[str, int]] = None,
            non_padded_namespaces: Iterable[str] = vocabulary.DEFAULT_NON_PADDED_NAMESPACES,
            pretrained_files: Optional[Dict[str, str]] = None,
            only_include_pretrained_words: bool = False,
            min_pretrained_embeddings: Dict[str, int] = None,
            padding_token: Optional[str] = vocabulary.DEFAULT_PADDING_TOKEN,
            oov_token: Optional[str] = vocabulary.DEFAULT_OOV_TOKEN,
    ) -> "Vocabulary":
        """
        Extension to manually fill gaps in missing 'feats_labels'.
        """
        # Load manually tokens from pretrained file (using different strategy
        # - only words add all embedding file, without checking if were seen
        # in any dataset.
        tokens_to_add = None
        if pretrained_files and "tokens" in pretrained_files:
            pretrained_set = set(vocabulary._read_pretrained_tokens(pretrained_files["tokens"]))
            tokens_to_add = {"tokens": list(pretrained_set)}
            pretrained_files = None

        vocab = super().from_instances(
            instances=instances,
            min_count=min_count,
            max_vocab_size=max_vocab_size,
            non_padded_namespaces=non_padded_namespaces,
            pretrained_files=pretrained_files,
            only_include_pretrained_words=only_include_pretrained_words,
            tokens_to_add=tokens_to_add,
            min_pretrained_embeddings=min_pretrained_embeddings,
            padding_token=padding_token,
            oov_token=oov_token
        )
        # Extending vocab with features that does not show up explicitly.
        # To know all features we need to read full dataset first.
        # Adding auxiliary '=None' feature for each category is needed
        # to perform classification.
        get_slices_if_not_provided(vocab)
        return vocab


def get_slices_if_not_provided(vocab: allen_data.Vocabulary):
    if hasattr(vocab, "slices"):
        return vocab.slices

    if "feats_labels" in vocab.get_namespaces():
        idx2token = vocab.get_index_to_token_vocabulary("feats_labels")
        for _, v in dict(idx2token).items():
            if v not in ["_", "__PAD__"]:
                empty_value = v.split("=")[0] + "=None"
                vocab.add_token_to_namespace(empty_value, "feats_labels")

        slices = {}
        for idx, name in vocab.get_index_to_token_vocabulary("feats_labels").items():
            # There are 2 types features: with (Case=Acc) or without assigment (None).
            # Here we group their indices by name (before assigment sign).
            name = name.split("=")[0]
            if name in slices:
                slices[name].append(idx)
            else:
                slices[name] = [idx]
        vocab.slices = slices
        return vocab.slices


@dataclass(init=False, repr=False)
class _Token(allen_data.Token):
    __slots__ = allen_data.Token.__slots__ + ['feats_']

    feats_: Optional[str]

    def __init__(self, text: str = None, idx: int = None, idx_end: int = None, lemma_: str = None, pos_: str = None,
                 tag_: str = None, dep_: str = None, ent_type_: str = None, text_id: int = None, type_id: int = None,
                 feats_: str = None) -> None:
        super().__init__(text, idx, idx_end, lemma_, pos_, tag_, dep_, ent_type_, text_id, type_id)
        self.feats_ = feats_



"""Finetuning train model wrapper."""
import os
from typing import List

from allennlp import data, models, common, training
from allennlp.commands import train
from allennlp.common import checks
from allennlp.common import util as common_util
from allennlp.training import util as training_util


@train.TrainModel.register("finetuning", constructor="from_partial_objects_finetuning")
class FinetuningTrainModel(train.TrainModel):
    """Class made only for finetuning, the only difference is saving vocab from concatenated
    (archive and current) datasets."""

    @classmethod
    def from_partial_objects_finetuning(
            cls,
            serialization_dir: str,
            local_rank: int,  # pylint: disable=unused-argument
            batch_weight_key: str,
            dataset_reader: data.DatasetReader,
            train_data_path: str,
            model: common.Lazy[models.Model],
            data_loader: common.Lazy[data.DataLoader],
            trainer: common.Lazy[training.Trainer],
            vocabulary: common.Lazy[data.Vocabulary] = None,
            datasets_for_vocab_creation: List[str] = None,
            validation_dataset_reader: data.DatasetReader = None,
            validation_data_path: str = None,
            validation_data_loader: common.Lazy[data.DataLoader] = None,
            test_data_path: str = None,
            evaluate_on_test: bool = False,
    ) -> "train.TrainModel":
        """
        This method is intended for use with our `FromParams` logic, to construct a `TrainModel`
        object from a config file passed to the `allennlp train` command.  The arguments to this
        method are the allowed top-level keys in a configuration file (except for the first three,
        which are obtained separately).

        You *could* use this outside of our `FromParams` logic if you really want to, but there
        might be easier ways to accomplish your goal than instantiating `Lazy` objects.  If you are
        writing your own training loop, we recommend that you look at the implementation of this
        method for inspiration and possibly some utility functions you can call, but you very likely
        should not use this method directly.

        The `Lazy` type annotations here are a mechanism for building dependencies to an object
        sequentially - the `TrainModel` object needs data, a model, and a trainer, but the model
        needs to see the data before it's constructed (to create a vocabulary) and the trainer needs
        the data and the model before it's constructed.  Objects that have sequential dependencies
        like this are labeled as `Lazy` in their type annotations, and we pass the missing
        dependencies when we call their `construct()` method, which you can see in the code below.

        # Parameters
        serialization_dir: `str`
            The directory where logs and model archives will be saved.
        local_rank: `int`
            The process index that is initialized using the GPU device id.
        batch_weight_key: `str`
            The name of metric used to weight the loss on a per-batch basis.
        dataset_reader: `DatasetReader`
            The `DatasetReader` that will be used for training and (by default) for validation.
        train_data_path: `str`
            The file (or directory) that will be passed to `dataset_reader.read()` to construct the
            training data.
        model: `Lazy[Model]`
            The model that we will train.  This is lazy because it depends on the `Vocabulary`;
            after constructing the vocabulary we call `model.construct(vocab=vocabulary)`.
        data_loader: `Lazy[DataLoader]`
            The data_loader we use to batch instances from the dataset reader at training and (by
            default) validation time. This is lazy because it takes a dataset in it's constructor.
        trainer: `Lazy[Trainer]`
            The `Trainer` that actually implements the training loop.  This is a lazy object because
            it depends on the model that's going to be trained.
        vocabulary: `Lazy[Vocabulary]`, optional (default=None)
            The `Vocabulary` that we will use to convert strings in the data to integer ids (and
            possibly set sizes of embedding matrices in the `Model`).  By default we construct the
            vocabulary from the instances that we read.
        datasets_for_vocab_creation: `List[str]`, optional (default=None)
            If you pass in more than one dataset but don't want to use all of them to construct a
            vocabulary, you can pass in this key to limit it.  Valid entries in the list are
            "train", "validation" and "test".
        validation_dataset_reader: `DatasetReader`, optional (default=None)
            If given, we will use this dataset reader for the validation data instead of
            `dataset_reader`.
        validation_data_path: `str`, optional (default=None)
            If given, we will use this data for computing validation metrics and early stopping.
        validation_data_loader: `Lazy[DataLoader]`, optional (default=None)
            If given, the data_loader we use to batch instances from the dataset reader at
            validation and test time. This is lazy because it takes a dataset in it's constructor.
        test_data_path: `str`, optional (default=None)
            If given, we will use this as test data.  This makes it available for vocab creation by
            default, but nothing else.
        evaluate_on_test: `bool`, optional (default=False)
            If given, we will evaluate the final model on this data at the end of training.  Note
            that we do not recommend using this for actual test data in every-day experimentation;
            you should only very rarely evaluate your model on actual test data.
        """

        datasets = training_util.read_all_datasets(
            train_data_path=train_data_path,
            dataset_reader=dataset_reader,
            validation_dataset_reader=validation_dataset_reader,
            validation_data_path=validation_data_path,
            test_data_path=test_data_path,
        )

        if datasets_for_vocab_creation:
            for key in datasets_for_vocab_creation:
                if key not in datasets:
                    raise checks.ConfigurationError(f"invalid 'dataset_for_vocab_creation' {key}")

        instance_generator = (
            instance
            for key, dataset in datasets.items()
            if not datasets_for_vocab_creation or key in datasets_for_vocab_creation
            for instance in dataset
        )

        vocabulary_ = vocabulary.construct(instances=instance_generator)
        if not vocabulary_:
            vocabulary_ = data.Vocabulary.from_instances(instance_generator)
        model_ = model.construct(vocab=vocabulary_)

        # Initializing the model can have side effect of expanding the vocabulary.
        # Save the vocab only in the master. In the degenerate non-distributed
        # case, we're trivially the master.
        if common_util.is_master():
            vocabulary_path = os.path.join(serialization_dir, "vocabulary")
            # Only difference compared to TrainModel!
            model_.vocab.save_to_files(vocabulary_path)

        for dataset in datasets.values():
            dataset.index_with(model_.vocab)

        data_loader_ = data_loader.construct(dataset=datasets["train"])
        validation_data = datasets.get("validation")
        if validation_data is not None:
            # Because of the way Lazy[T] works, we can't check it's existence
            # _before_ we've tried to construct it. It returns None if it is not
            # present, so we try to construct it first, and then afterward back off
            # to the data_loader configuration used for training if it returns None.
            validation_data_loader_ = validation_data_loader.construct(dataset=validation_data)
            if validation_data_loader_ is None:
                validation_data_loader_ = data_loader.construct(dataset=validation_data)
        else:
            validation_data_loader_ = None

        test_data = datasets.get("test")
        if test_data is not None:
            test_data_loader = validation_data_loader.construct(dataset=test_data)
            if test_data_loader is None:
                test_data_loader = data_loader.construct(dataset=test_data)
        else:
            test_data_loader = None

        # We don't need to pass serialization_dir and local_rank here, because they will have been
        # passed through the trainer by from_params already, because they were keyword arguments to
        # construct this class in the first place.
        trainer_ = trainer.construct(
            model=model_, data_loader=data_loader_, validation_data_loader=validation_data_loader_,
        )

        return cls(
            serialization_dir=serialization_dir,
            model=model_,
            trainer=trainer_,
            evaluation_data_loader=test_data_loader,
            evaluate_on_test=evaluate_on_test,
            batch_weight_key=batch_weight_key,
        )



from typing import Optional, Dict, Any, List, Tuple

from allennlp import data
from allennlp.data import token_indexers, tokenizers


@data.TokenIndexer.register("pretrained_transformer_mismatched_fixed")
class PretrainedTransformerMismatchedIndexer(token_indexers.PretrainedTransformerMismatchedIndexer):
    """TODO(mklimasz) Remove during next allennlp update, fixed on allennlp master."""

    def __init__(self, model_name: str, namespace: str = "tags", max_length: int = None,
                 tokenizer_kwargs: Optional[Dict[str, Any]] = None, **kwargs) -> None:
        # The matched version v.s. mismatched
        super().__init__(model_name, namespace, max_length, tokenizer_kwargs, **kwargs)
        self._matched_indexer = PretrainedTransformerIndexer(
            model_name,
            namespace=namespace,
            max_length=max_length,
            tokenizer_kwargs=tokenizer_kwargs,
            **kwargs,
        )
        self._allennlp_tokenizer = self._matched_indexer._allennlp_tokenizer
        self._tokenizer = self._matched_indexer._tokenizer
        self._num_added_start_tokens = self._matched_indexer._num_added_start_tokens
        self._num_added_end_tokens = self._matched_indexer._num_added_end_tokens


class PretrainedTransformerIndexer(token_indexers.PretrainedTransformerIndexer):

    def __init__(
            self,
            model_name: str,
            namespace: str = "tags",
            max_length: int = None,
            tokenizer_kwargs: Optional[Dict[str, Any]] = None,
            **kwargs,
    ) -> None:
        super().__init__(model_name, namespace, max_length, tokenizer_kwargs, **kwargs)
        self._namespace = namespace
        self._allennlp_tokenizer = PretrainedTransformerTokenizer(
            model_name, tokenizer_kwargs=tokenizer_kwargs
        )
        self._tokenizer = self._allennlp_tokenizer.tokenizer
        self._added_to_vocabulary = False

        self._num_added_start_tokens = len(self._allennlp_tokenizer.single_sequence_start_tokens)
        self._num_added_end_tokens = len(self._allennlp_tokenizer.single_sequence_end_tokens)

        self._max_length = max_length
        if self._max_length is not None:
            num_added_tokens = len(self._allennlp_tokenizer.tokenize("a")) - 1
            self._effective_max_length = (  # we need to take into account special tokens
                    self._max_length - num_added_tokens
            )
            if self._effective_max_length <= 0:
                raise ValueError(
                    "max_length needs to be greater than the number of special tokens inserted."
                )


class PretrainedTransformerTokenizer(tokenizers.PretrainedTransformerTokenizer):

    def _intra_word_tokenize(
            self, string_tokens: List[str]
    ) -> Tuple[List[data.Token], List[Optional[Tuple[int, int]]]]:
        tokens: List[data.Token] = []
        offsets: List[Optional[Tuple[int, int]]] = []
        for token_string in string_tokens:
            wordpieces = self.tokenizer.encode_plus(
                token_string,
                add_special_tokens=False,
                return_tensors=None,
                return_offsets_mapping=False,
                return_attention_mask=False,
            )
            wp_ids = wordpieces["input_ids"]

            if len(wp_ids) > 0:
                offsets.append((len(tokens), len(tokens) + len(wp_ids) - 1))
                tokens.extend(
                    data.Token(text=wp_text, text_id=wp_id)
                    for wp_id, wp_text in zip(wp_ids, self.tokenizer.convert_ids_to_tokens(wp_ids))
                )
            else:
                offsets.append(None)
        return tokens, offsets



import logging
from typing import Dict, Optional, List

import torch
from allennlp import models, common
from allennlp.data import dataloader
from allennlp.training import optimizers

logger = logging.getLogger(__name__)


class NullTensorboardWriter(common.FromParams):

    def log_batch(
            self,
            model: models.Model,
            optimizer: optimizers.Optimizer,
            batch_grad_norm: Optional[float],
            metrics: Dict[str, float],
            batch_group: List[List[dataloader.TensorDict]],
            param_updates: Optional[Dict[str, torch.Tensor]],
    ) -> None:
        pass

    def reset_epoch(self) -> None:
        pass

    def should_log_this_batch(self) -> bool:
        return False

    def should_log_histograms_this_batch(self) -> bool:
        return False

    def add_train_scalar(self, name: str, value: float, timestep: int = None) -> None:
        pass

    def add_train_histogram(self, name: str, values: torch.Tensor) -> None:
        pass

    def add_validation_scalar(self, name: str, value: float, timestep: int = None) -> None:
        pass

    def log_parameter_and_gradient_statistics(self, model: models.Model, batch_grad_norm: float) -> None:
        pass

    def log_learning_rates(self, model: models.Model, optimizer: torch.optim.Optimizer):
        pass

    def log_histograms(self, model: models.Model) -> None:
        pass

    def log_gradient_updates(self, model: models.Model, param_updates: Dict[str, torch.Tensor]) -> None:
        pass

    def log_metrics(
        self,
        train_metrics: dict,
        val_metrics: dict = None,
        epoch: int = None,
        log_to_console: bool = False,
    ) -> None:
        metric_names = set(train_metrics.keys())
        if val_metrics is not None:
            metric_names.update(val_metrics.keys())
        val_metrics = val_metrics or {}

        if log_to_console:
            dual_message_template = "%s |  %8.3f  |  %8.3f"
            no_val_message_template = "%s |  %8.3f  |  %8s"
            no_train_message_template = "%s |  %8s  |  %8.3f"
            header_template = "%s |  %-10s"
            name_length = max(len(x) for x in metric_names)
            logger.info(header_template, "Training".rjust(name_length + 13), "Validation")

            for name in metric_names:
                train_metric = train_metrics.get(name)
                val_metric = val_metrics.get(name)
                if val_metric is not None and train_metric is not None:
                    logger.info(
                        dual_message_template, name.ljust(name_length), train_metric, val_metric
                    )
                elif val_metric is not None:
                    logger.info(no_train_message_template, name.ljust(name_length), "N/A", val_metric)
                elif train_metric is not None:
                    logger.info(no_val_message_template, name.ljust(name_length), train_metric, "N/A")

    def enable_activation_logging(self, model: models.Model) -> None:
        pass

    def log_activation_histogram(self, outputs, log_prefix: str) -> None:
        pass

    def close(self) -> None:
        pass



"""Main COMBO model."""
from typing import Optional, Dict, Any, List

import torch
from allennlp import data, modules, models as allen_models, nn as allen_nn
from allennlp.modules import text_field_embedders
from allennlp.nn import util
from overrides import overrides

from unidic_combo.models import base
from unidic_combo.utils import metrics


@allen_models.Model.register("semantic_multitask")
class SemanticMultitaskModel(allen_models.Model):
    """Main COMBO model."""

    def __init__(self,
                 vocab: data.Vocabulary,
                 loss_weights: Dict[str, float],
                 text_field_embedder: text_field_embedders.TextFieldEmbedder,
                 seq_encoder: modules.Seq2SeqEncoder,
                 use_sample_weight: bool = True,
                 lemmatizer: Optional[base.Predictor] = None,
                 upos_tagger: Optional[base.Predictor] = None,
                 xpos_tagger: Optional[base.Predictor] = None,
                 semantic_relation: Optional[base.Predictor] = None,
                 morphological_feat: Optional[base.Predictor] = None,
                 dependency_relation: Optional[base.Predictor] = None,
                 regularizer: allen_nn.RegularizerApplicator = None) -> None:
        super().__init__(vocab, regularizer)
        self.text_field_embedder = text_field_embedder
        self.loss_weights = loss_weights
        self.use_sample_weight = use_sample_weight
        self.seq_encoder = seq_encoder
        self.lemmatizer = lemmatizer
        self.upos_tagger = upos_tagger
        self.xpos_tagger = xpos_tagger
        self.semantic_relation = semantic_relation
        self.morphological_feat = morphological_feat
        self.dependency_relation = dependency_relation
        self._head_sentinel = torch.nn.Parameter(torch.randn([1, 1, self.seq_encoder.get_output_dim()]))
        self.scores = metrics.SemanticMetrics()
        self._partial_losses = None

    @overrides
    def forward(self,
                sentence: Dict[str, Dict[str, torch.Tensor]],
                metadata: List[Dict[str, Any]],
                upostag: torch.Tensor = None,
                xpostag: torch.Tensor = None,
                lemma: Dict[str, Dict[str, torch.Tensor]] = None,
                feats: torch.Tensor = None,
                head: torch.Tensor = None,
                deprel: torch.Tensor = None,
                semrel: torch.Tensor = None, ) -> Dict[str, torch.Tensor]:

        # Prepare masks
        char_mask: torch.BoolTensor = sentence["char"]["token_characters"] > 0
        word_mask = util.get_text_field_mask(sentence)

        # If enabled weight samples loss by log(sentence_length)
        sample_weights = word_mask.sum(-1).float().log() if self.use_sample_weight else None

        encoder_input = self.text_field_embedder(sentence, char_mask=char_mask)
        encoder_emb = self.seq_encoder(encoder_input, word_mask)

        batch_size, _, encoding_dim = encoder_emb.size()

        # Concatenate the head sentinel (ROOT) onto the sentence representation.
        head_sentinel = self._head_sentinel.expand(batch_size, 1, encoding_dim)
        encoder_emb = torch.cat([head_sentinel, encoder_emb], 1)
        word_mask = torch.cat([word_mask.new_ones((batch_size, 1)), word_mask], 1)

        upos_output = self._optional(self.upos_tagger,
                                     encoder_emb[:, 1:],
                                     mask=word_mask[:, 1:],
                                     labels=upostag,
                                     sample_weights=sample_weights)
        xpos_output = self._optional(self.xpos_tagger,
                                     encoder_emb[:, 1:],
                                     mask=word_mask[:, 1:],
                                     labels=xpostag,
                                     sample_weights=sample_weights)
        semrel_output = self._optional(self.semantic_relation,
                                       encoder_emb[:, 1:],
                                       mask=word_mask[:, 1:],
                                       labels=semrel,
                                       sample_weights=sample_weights)
        morpho_output = self._optional(self.morphological_feat,
                                       encoder_emb[:, 1:],
                                       mask=word_mask[:, 1:],
                                       labels=feats,
                                       sample_weights=sample_weights)
        lemma_output = self._optional(self.lemmatizer,
                                      (encoder_emb[:, 1:], sentence.get("char").get("token_characters")
                                       if sentence.get("char") else None),
                                      mask=word_mask[:, 1:],
                                      labels=lemma.get("char").get("token_characters") if lemma else None,
                                      sample_weights=sample_weights)
        parser_output = self._optional(self.dependency_relation,
                                       encoder_emb,
                                       returns_tuple=True,
                                       mask=word_mask,
                                       labels=(deprel, head),
                                       sample_weights=sample_weights)
        relations_pred, head_pred = parser_output["prediction"]
        output = {
            "upostag": upos_output["prediction"],
            "xpostag": xpos_output["prediction"],
            "semrel": semrel_output["prediction"],
            "feats": morpho_output["prediction"],
            "lemma": lemma_output["prediction"],
            "head": head_pred,
            "deprel": relations_pred,
            "sentence_embedding": torch.max(encoder_emb[:, 1:], dim=1)[0],
        }

        if self._has_labels([upostag, xpostag, lemma, feats, head, deprel, semrel]):

            # Feats mapping
            if self.morphological_feat:
                mapped_gold_labels = []
                for _, cat_indices in self.morphological_feat.slices.items():
                    mapped_gold_labels.append(feats[:, :, cat_indices].argmax(dim=-1))

                feats = torch.stack(mapped_gold_labels, dim=-1)

            labels = {
                "upostag": upostag,
                "xpostag": xpostag,
                "semrel": semrel,
                "feats": feats,
                "lemma": lemma.get("char").get("token_characters") if lemma else None,
                "head": head,
                "deprel": deprel,
            }
            self.scores(output, labels, word_mask[:, 1:])
            relations_loss, head_loss = parser_output["loss"]
            losses = {
                "upostag_loss": upos_output["loss"],
                "xpostag_loss": xpos_output["loss"],
                "semrel_loss": semrel_output["loss"],
                "feats_loss": morpho_output["loss"],
                "lemma_loss": lemma_output["loss"],
                "head_loss": head_loss,
                "deprel_loss": relations_loss,
                # Cycle loss is only for the metrics purposes.
                "cycle_loss": parser_output.get("cycle_loss")
            }
            self._partial_losses = losses.copy()
            losses["loss"] = self._calculate_loss(losses)
            output.update(losses)

        return self._clean(output)

    @staticmethod
    def _has_labels(labels):
        return any(x is not None for x in labels)

    def _calculate_loss(self, output):
        losses = []
        for name, value in self.loss_weights.items():
            if output.get(f"{name}_loss"):
                losses.append(output[f"{name}_loss"] * value)
        return torch.stack(losses).sum()

    @staticmethod
    def _optional(callable_model: Optional[torch.nn.Module],
                  *args,
                  returns_tuple: bool = False,
                  **kwargs):
        if callable_model:
            return callable_model(*args, **kwargs)
        if returns_tuple:
            return {"prediction": (None, None), "loss": (None, None)}
        return {"prediction": None, "loss": None}

    @staticmethod
    def _clean(output):
        for k, v in dict(output).items():
            if v is None:
                del output[k]
        return output

    @overrides
    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = self.scores.get_metric(reset)
        if self._partial_losses:
            losses = self._clean(self._partial_losses)
            losses = {f"partial_loss/{k}": v.detach().item() for k, v in losses.items()}
            metrics.update(losses)
        return metrics



import torch
import torch.nn.functional as F


def masked_cross_entropy(pred: torch.Tensor, true: torch.Tensor, mask: torch.BoolTensor) -> torch.Tensor:
    pred = pred + (mask.float().unsqueeze(-1) + 1e-45).log()
    return F.cross_entropy(pred, true, reduction="none") * mask



"""Embeddings."""
from typing import Optional, Dict, Any

import torch
import torch.nn as nn
from allennlp import nn as allen_nn, data, modules
from allennlp.modules import token_embedders
from allennlp.nn import util
from overrides import overrides

from unidic_combo.models import base, dilated_cnn


@token_embedders.TokenEmbedder.register("char_embeddings")
@token_embedders.TokenEmbedder.register("char_embeddings_from_config", constructor="from_config")
class CharacterBasedWordEmbeddings(token_embedders.TokenEmbedder):
    """Character-based word embeddings."""

    def __init__(self,
                 num_embeddings: int,
                 embedding_dim: int,
                 dilated_cnn_encoder: dilated_cnn.DilatedCnnEncoder):
        super().__init__()
        self.char_embed = nn.Embedding(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        self.dilated_cnn_encoder = modules.TimeDistributed(dilated_cnn_encoder)
        self.output_dim = embedding_dim

    def forward(self,
                x: torch.Tensor,
                char_mask: Optional[torch.BoolTensor] = None) -> torch.Tensor:
        if char_mask is None:
            char_mask = x.new_ones(x.size())

        x = self.char_embed(x)
        x = x * char_mask.unsqueeze(-1).float()
        x = self.dilated_cnn_encoder(x.transpose(2, 3))
        return torch.max(x, dim=-1)[0]

    @overrides
    def get_output_dim(self) -> int:
        return self.output_dim

    @classmethod
    def from_config(cls,
                    embedding_dim: int,
                    vocab: data.Vocabulary,
                    dilated_cnn_encoder: dilated_cnn.DilatedCnnEncoder,
                    vocab_namespace: str = "token_characters"):
        assert vocab_namespace in vocab.get_namespaces()
        return cls(
            embedding_dim=embedding_dim,
            num_embeddings=vocab.get_vocab_size(vocab_namespace),
            dilated_cnn_encoder=dilated_cnn_encoder
        )


@token_embedders.TokenEmbedder.register("embeddings_projected")
class ProjectedWordEmbedder(token_embedders.Embedding):
    """Word embeddings."""

    def __init__(self,
                 embedding_dim: int,
                 num_embeddings: int = None,
                 weight: torch.FloatTensor = None,
                 padding_index: int = None,
                 trainable: bool = True,
                 max_norm: float = None,
                 norm_type: float = 2.0,
                 scale_grad_by_freq: bool = False,
                 sparse: bool = False,
                 vocab_namespace: str = "tokens",
                 pretrained_file: str = None,
                 vocab: data.Vocabulary = None,
                 projection_layer: Optional[base.Linear] = None):
        super().__init__(
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            weight=weight,
            padding_index=padding_index,
            trainable=trainable,
            max_norm=max_norm,
            norm_type=norm_type,
            scale_grad_by_freq=scale_grad_by_freq,
            sparse=sparse,
            vocab_namespace=vocab_namespace,
            pretrained_file=pretrained_file,
            vocab=vocab
        )
        self._projection = projection_layer
        self.output_dim = embedding_dim if projection_layer is None else projection_layer.out_features

    @overrides
    def get_output_dim(self) -> int:
        return self.output_dim


@token_embedders.TokenEmbedder.register("transformers_word_embeddings")
class TransformersWordEmbedder(token_embedders.PretrainedTransformerMismatchedEmbedder):
    """
    Transformers word embeddings as last hidden state + optional projection layers.

    Tested with Bert (but should work for other models as well).
    """

    def __init__(self,
                 model_name: str,
                 projection_dim: int,
                 projection_activation: Optional[allen_nn.Activation] = lambda x: x,
                 projection_dropout_rate: Optional[float] = 0.0,
                 freeze_transformer: bool = True,
                 tokenizer_kwargs: Optional[Dict[str, Any]] = None,
                 transformer_kwargs: Optional[Dict[str, Any]] = None):
        super().__init__(model_name, tokenizer_kwargs=tokenizer_kwargs, transformer_kwargs=transformer_kwargs)
        self.freeze_transformer = freeze_transformer
        if self.freeze_transformer:
            self._matched_embedder.eval()
            for param in self._matched_embedder.parameters():
                param.requires_grad = False
        if projection_dim:
            self.projection_layer = base.Linear(in_features=super().get_output_dim(),
                                                out_features=projection_dim,
                                                dropout_rate=projection_dropout_rate,
                                                activation=projection_activation)
            self.output_dim = projection_dim
        else:
            self.projection_layer = None
            self.output_dim = super().get_output_dim()

    @overrides
    def forward(
            self,
            token_ids: torch.LongTensor,
            mask: torch.BoolTensor,
            offsets: torch.LongTensor,
            wordpiece_mask: torch.BoolTensor,
            type_ids: Optional[torch.LongTensor] = None,
            segment_concat_mask: Optional[torch.BoolTensor] = None,
    ) -> torch.Tensor:  # type: ignore
        x = super().forward(token_ids, mask, offsets, wordpiece_mask, type_ids, segment_concat_mask)
        if self.projection_layer:
            x = self.projection_layer(x)
        return x

    @overrides
    def get_output_dim(self):
        return self.output_dim

    @overrides
    def train(self, mode: bool):
        if self.freeze_transformer:
            self.projection_layer.train(mode)
        else:
            super().train(mode)

    @overrides
    def eval(self):
        if self.freeze_transformer:
            self.projection_layer.eval()
        else:
            super().eval()


@token_embedders.TokenEmbedder.register("feats_embedding")
class FeatsTokenEmbedder(token_embedders.Embedding):

    def __init__(self,
                 embedding_dim: int,
                 num_embeddings: int = None,
                 weight: torch.FloatTensor = None,
                 padding_index: int = None,
                 trainable: bool = True,
                 max_norm: float = None,
                 norm_type: float = 2.0,
                 scale_grad_by_freq: bool = False,
                 sparse: bool = False,
                 vocab_namespace: str = "feats",
                 pretrained_file: str = None,
                 vocab: data.Vocabulary = None):
        super().__init__(
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            weight=weight,
            padding_index=padding_index,
            trainable=trainable,
            max_norm=max_norm,
            norm_type=norm_type,
            scale_grad_by_freq=scale_grad_by_freq,
            sparse=sparse,
            vocab_namespace=vocab_namespace,
            pretrained_file=pretrained_file,
            vocab=vocab
        )

    def forward(self, tokens: torch.Tensor) -> torch.Tensor:
        # (batch_size, sentence_length, features_vocab_length)
        mask = (tokens > 0).float()
        # (batch_size, sentence_length, features_vocab_length, embedding_dim)
        x = super().forward(tokens)
        # (batch_size, sentence_length, embedding_dim)
        return x.sum(dim=-2) / (
            (mask.sum(dim=-1) + util.tiny_value_of_dtype(mask.dtype)).unsqueeze(dim=-1)
        )



import os,setuptools,subprocess
if os.path.isfile("dummy.tar.gz"):
  package_data={"unidic_combo":["./mkmodel.sh"]}
else:
  subprocess.run(["unidic_combo/mkmodel.sh"])
  package_data={"unidic_combo":["./mkmodel.sh","./combo-*.tar.gz"]}
with open("README.md","r",encoding="utf-8") as r:
  long_description=r.read()
URL="https://github.com/KoichiYasuoka/UniDic-COMBO"

setuptools.setup(
  name="unidic_combo",
  version="0.8.5",
  description="UniDic2UD + COMBO-pytorch wrapper for spaCy",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url=URL,
  author="Koichi Yasuoka",
  author_email="yasuoka@kanji.zinbun.kyoto-u.ac.jp",
  license="GPL",
  keywords="NLP Japanese spaCy",
  packages=setuptools.find_packages(),
  install_requires=[
    "unidic2ud>=2.6.6",
    "spacy>=2.2.2",
    "allennlp>=1.2.0",
    "torch>=1.6.0",
    "absl-py>=0.9.0",
    "conllu>=2.3.2",
    "dataclasses-json>=0.5.2",
    "requests>=2.23.0",
    "overrides>=3.1.0",
    "fugashi>=1.0.1",
    "ipadic>=1.0.0"
  ],
  python_requires=">=3.6",
  package_data=package_data,
  classifiers=[
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
    "Topic :: Text Processing :: Linguistic",
    "Natural Language :: Japanese"
  ],
  project_urls={
    "COMBO-pytorch":"https://github.com/ipipan/combo",
    "Source":URL,
    "Tracker":URL+"/issues",
  }
)



"""Main entry point."""
import logging
import os
import tempfile
from typing import Dict

import torch
from absl import app
from absl import flags
from allennlp import common, models, predictors
from allennlp.commands import train, predict as allen_predict
from allennlp.common import checks as allen_checks, util
from allennlp.models import archival

from unidic_combo import predict
from unidic_combo.data import api, dataset
from unidic_combo.utils import checks

logger = logging.getLogger(__name__)
_FEATURES = ["token", "char", "upostag", "xpostag", "lemma", "feats"]
_TARGETS = ["deprel", "feats", "head", "lemma", "upostag", "xpostag", "semrel", "sent"]

FLAGS = flags.FLAGS
flags.DEFINE_enum(name="mode", default=None, enum_values=["train", "predict"],
                  help="Specify COMBO mode: train or predict")

# Common flags
flags.DEFINE_integer(name="cuda_device", default=-1,
                     help="Cuda device id (default -1 cpu)")
flags.DEFINE_string(name="output_file", default="output.log",
                    help="Predictions result file.")

# Training flags
flags.DEFINE_list(name="training_data_path", default="./tests/fixtures/example.conllu",
                  help="Training data path(s)")
flags.DEFINE_list(name="validation_data_path", default="",
                  help="Validation data path(s)")
flags.DEFINE_string(name="pretrained_tokens", default="",
                    help="Pretrained tokens embeddings path")
flags.DEFINE_integer(name="embedding_dim", default=300,
                     help="Embeddings dim")
flags.DEFINE_integer(name="num_epochs", default=400,
                     help="Epochs num")
flags.DEFINE_integer(name="word_batch_size", default=2500,
                     help="Minimum words in batch")
flags.DEFINE_string(name="pretrained_transformer_name", default="",
                    help="Pretrained transformer model name (see transformers from HuggingFace library for list of "
                         "available models) for transformers based embeddings.")
flags.DEFINE_list(name="features", default=["token", "char"],
                  help=f"Features used to train model (required 'token' and 'char'). Possible values: {_FEATURES}.")
flags.DEFINE_list(name="targets", default=["deprel", "feats", "head", "lemma", "upostag", "xpostag"],
                  help=f"Targets of the model (required `deprel` and `head`). Possible values: {_TARGETS}.")
flags.DEFINE_string(name="serialization_dir", default=None,
                    help="Model serialization directory (default - system temp dir).")
flags.DEFINE_boolean(name="tensorboard", default=False,
                     help="When provided model will log tensorboard metrics.")

# Finetune after training flags
flags.DEFINE_list(name="finetuning_training_data_path", default="",
                  help="Training data path(s)")
flags.DEFINE_list(name="finetuning_validation_data_path", default="",
                  help="Validation data path(s)")
flags.DEFINE_string(name="config_path", default="config.template.jsonnet",
                    help="Config file path.")

# Test after training flags
flags.DEFINE_string(name="test_path", default=None,
                    help="Test path file.")

# Experimental
flags.DEFINE_boolean(name="use_pure_config", default=False,
                     help="Ignore ext flags (experimental).")

# Prediction flags
flags.DEFINE_string(name="model_path", default=None,
                    help="Pretrained model path.")
flags.DEFINE_string(name="input_file", default=None,
                    help="File to predict path")
flags.DEFINE_boolean(name="conllu_format", default=True,
                     help="Prediction based on conllu format (instead of raw text).")
flags.DEFINE_integer(name="batch_size", default=1,
                     help="Prediction batch size.")
flags.DEFINE_boolean(name="silent", default=True,
                     help="Silent prediction to file (without printing to console).")
flags.DEFINE_enum(name="predictor_name", default="semantic-multitask-predictor-spacy",
                  enum_values=["semantic-multitask-predictor", "semantic-multitask-predictor-spacy"],
                  help="Use predictor with whitespace or spacy tokenizer.")


def run(_):
    """Run model."""
    # Imports are required to make Registrable modules visible without passing parameter
    util.import_module_and_submodules("unidic_combo.commands")
    util.import_module_and_submodules("unidic_combo.models")
    util.import_module_and_submodules("unidic_combo.training")

    if FLAGS.mode == "train":
        checks.file_exists(FLAGS.config_path)
        params = common.Params.from_file(FLAGS.config_path, ext_vars=_get_ext_vars())
        model_params = params.get("model").as_ordered_dict()
        serialization_dir = tempfile.mkdtemp(prefix="allennlp", dir=FLAGS.serialization_dir)
        model = train.train_model(params, serialization_dir=serialization_dir, file_friendly_logging=True)
        logger.info(f"Training model stored in: {serialization_dir}")

        if FLAGS.finetuning_training_data_path:
            for f in FLAGS.finetuning_training_data_path:
                checks.file_exists(f)

            # Loading will be performed from stored model.tar.gz
            del model
            if torch.cuda.is_available():
                torch.cuda.empty_cache()

            params = common.Params.from_file(FLAGS.config_path, ext_vars=_get_ext_vars(finetuning=True))
            # Replace model definition with pretrained archive
            params["model"] = {
                "type": "from_archive",
                "archive_file": serialization_dir + "/model.tar.gz",
            }
            serialization_dir = tempfile.mkdtemp(prefix="allennlp", suffix="-finetuning", dir=FLAGS.serialization_dir)
            model = train.train_model(params.duplicate(), serialization_dir=serialization_dir,
                                      file_friendly_logging=True)

            # Make finetuning model serialization independent from training serialization
            # Storing model definition instead of archive
            params["model"] = model_params
            params.to_file(os.path.join(serialization_dir, archival.CONFIG_NAME))
            archival.archive_model(serialization_dir)

            logger.info(f"Finetuned model stored in: {serialization_dir}")

        if FLAGS.test_path and FLAGS.output_file:
            checks.file_exists(FLAGS.test_path)
            params = common.Params.from_file(FLAGS.config_path, ext_vars=_get_ext_vars())["dataset_reader"]
            params.pop("type")
            dataset_reader = dataset.UniversalDependenciesDatasetReader.from_params(params)
            predictor = predict.SemanticMultitaskPredictor(
                model=model,
                dataset_reader=dataset_reader
            )
            test_trees = dataset_reader.read(FLAGS.test_path)
            with open(FLAGS.output_file, "w") as file:
                for tree in test_trees:
                    file.writelines(api.sentence2conllu(predictor.predict_instance(tree),
                                                        keep_semrel=dataset_reader.use_sem).serialize())
    else:
        use_dataset_reader = FLAGS.conllu_format
        predictor = _get_predictor()
        if FLAGS.input_file == "-":
            use_dataset_reader = False
            predictor.without_sentence_embedding = True
        if use_dataset_reader:
            predictor.line_to_conllu = True
        if FLAGS.silent:
            logging.getLogger("allennlp.common.params").disabled = True
        manager = allen_predict._PredictManager(
            predictor,
            FLAGS.input_file,
            FLAGS.output_file,
            FLAGS.batch_size,
            not FLAGS.silent,
            use_dataset_reader,
        )
        manager.run()


def _get_predictor() -> predictors.Predictor:
    allen_checks.check_for_gpu(FLAGS.cuda_device)
    checks.file_exists(FLAGS.model_path)
    archive = models.load_archive(
        FLAGS.model_path,
        cuda_device=FLAGS.cuda_device,
    )

    return predictors.Predictor.from_archive(
        archive, FLAGS.predictor_name
    )


def _get_ext_vars(finetuning: bool = False) -> Dict:
    if FLAGS.use_pure_config:
        return {}
    return {
        "training_data_path": (
            ",".join(FLAGS.training_data_path if not finetuning else FLAGS.finetuning_training_data_path)),
        "validation_data_path": (
            ",".join(FLAGS.validation_data_path if not finetuning else FLAGS.finetuning_validation_data_path)),
        "pretrained_tokens": FLAGS.pretrained_tokens,
        "pretrained_transformer_name": FLAGS.pretrained_transformer_name,
        "features": " ".join(FLAGS.features),
        "targets": " ".join(FLAGS.targets),
        "type": "finetuning" if finetuning else "default",
        "embedding_dim": str(FLAGS.embedding_dim),
        "cuda_device": str(FLAGS.cuda_device),
        "num_epochs": str(FLAGS.num_epochs),
        "word_batch_size": str(FLAGS.word_batch_size),
        "use_tensorboard": str(FLAGS.tensorboard),
    }


def main():
    """Parse flags."""
    flags.register_validator(
        "features",
        lambda values: all(
            value in _FEATURES for value in values),
        message="Flags --features contains unknown value(s)."
    )
    flags.register_validator(
        "mode",
        lambda value: value is not None,
        message="Flag --mode must be set with either `predict` or `train` value")
    flags.register_validator(
        "targets",
        lambda values: all(
            value in _TARGETS for value in values),
        message="Flag --targets contains unknown value(s)."
    )
    app.run(run)


if __name__ == "__main__":
    main()
